/*
 * http://www.pabr.org/bricks/brickhid
 * Main program
 *
 * Copyright (c) 2014 pabr@pabr.org
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/errno.h>

#include "ev3wifi.h"
#include "bthid.h"

void fatal(const char *msg) { perror(msg); exit(1); }

#define MAX_MOTORS 4
#define MAX_AXES 32

struct {
  int origin;
  int deadzone;
  enum { MODE_DISABLE=1, MODE_LINEAR, MODE_QUIT, MODE_HOME } mode;
} axes[MAX_AXES];

int naxes;

struct {
  char mode;            // 'p'osition or 's'peed or 'd'isable
  int gain;             // Servo gain for mode 'p'
  int maxspeed;         // Maximum speed
  int homespeed;        // Speed while homing
  int coefs[MAX_AXES];  // Linear mixing
  int offset;           // Added to output
} motors[MAX_MOTORS];

void parse_config(const char *configfile) {
  FILE *f = fopen(configfile, "r");
  if ( ! f ) fatal(configfile);
  char line[256], mode[16];
  while ( fgets(line,sizeof(line),f) ) {
    if      ( line[0] == '#' ) ;
    else if ( line[0] == 0 ) ;
    else if ( sscanf(line, "motor_mode %c %c %c %c",
		     &motors[0].mode, &motors[1].mode,
		     &motors[2].mode, &motors[3].mode) == 4 ) ;
    else if ( sscanf(line, "motor_maxspeed %d %d %d %d",
		     &motors[0].maxspeed, &motors[1].maxspeed,
		     &motors[2].maxspeed, &motors[3].maxspeed) == 4 ) ;
    else if ( sscanf(line, "motor_homespeed %d %d %d %d",
		     &motors[0].homespeed, &motors[1].homespeed,
		     &motors[2].homespeed, &motors[3].homespeed) == 4 ) ;
    else if ( sscanf(line, "motor_gain %d %d %d %d",
		     &motors[0].gain, &motors[1].gain,
		     &motors[2].gain, &motors[3].gain) == 4 ) ;
    else if ( sscanf(line, "motor_offset %d %d %d %d",
		     &motors[0].offset, &motors[1].offset,
		     &motors[2].offset, &motors[3].offset) == 4 ) ;
    else if ( naxes<MAX_AXES &&
	      sscanf(line, "axis %*s  %d  %d  %d %d %d %d %15s",
		     &axes[naxes].origin, &axes[naxes].deadzone,
		     &motors[0].coefs[naxes], &motors[1].coefs[naxes],
		     &motors[2].coefs[naxes], &motors[3].coefs[naxes],
		     mode) == 7 ) {
      if      ( ! strcmp(mode,"disable") ) axes[naxes].mode = MODE_DISABLE;
      else if ( ! strcmp(mode,"linear") )  axes[naxes].mode = MODE_LINEAR;
      else if ( ! strcmp(mode,"home") )    axes[naxes].mode = MODE_HOME;
      else if ( ! strcmp(mode,"quit") )    axes[naxes].mode = MODE_QUIT;
      else fatal(line);
      ++naxes;
    }
    else fatal(line);
  }
}

void usage() {
  fprintf(stderr, "usage: brickhid [configfile]\n");
  exit(1);
}

void home(struct ev3_conn *ev3) {
  int nos = 0;
  for ( int m=0; m<MAX_MOTORS; ++m ) {
    if ( motors[m].mode == 'p' ) {
      ev3_motor_speed(ev3, 1<<m, motors[m].homespeed);
      nos |= 1<<m;
    }
  }
  ev3_motor_start(ev3, nos);
  sleep(3);
  ev3_motor_stop(ev3, nos, BRAKE);
  sleep(1);
  ev3_motor_clr_count(ev3, nos);
}

int main(int argc, char *argv[]) {
  const char *configfile = NULL;
  for ( int i=1; i<argc; ++i ) {
    configfile = argv[i];
  }
  if ( ! configfile ) usage();
  parse_config(configfile);
  
  struct ev3_conn *ev3 = ev3_connect(NULL, NULL);
  ev3_sound_tone(ev3, 50, 440, 100);
  ev3_motor_speed(ev3, MOTA|MOTB|MOTC|MOTD, 0);
  ev3_motor_start(ev3, MOTA|MOTB|MOTC|MOTD);
  ev3_motor_clr_count(ev3, MOTA|MOTB|MOTC|MOTD);

  struct bthid_conn *hid = bthid_accept();
  ev3_sound_tone(ev3, 50, 880, 100);

  int quit_requested = 0;
  do {
    int axvals[MAX_AXES];
    int nax=0, nr, count=0;
    while ( (nr=bthid_recv_axes(hid, axvals, MAX_AXES)) > 0 ) {
      nax = nr;
      ++count;
    }
    if ( nr == 0 ) fatal("Bluetooth connection closed");
    if ( errno != EAGAIN ) fatal("Bluetooth connection lost");
    if ( ! count ) continue;
    fprintf(stderr, "(%d)", count);
    // Special functions
    for ( int a=0; a<naxes; ++a ) {
      if ( axvals[a] && axes[a].mode==MODE_QUIT ) quit_requested = 1;
      if ( axvals[a] && axes[a].mode==MODE_HOME ) home(ev3);
    }
    // Linear mixing
    for ( int m=0; m<MAX_MOTORS; ++m ) {
      int v = motors[m].offset;
      for ( int a=0; a<nax && a<naxes; ++a ) {
	if ( axes[a].mode != MODE_LINEAR ) continue;
	int axval = axvals[a] - axes[a].origin;
	if ( axval > 0 ) {
	  axval -= axes[a].deadzone;
	  if ( axval < 0 ) axval = 0;
	} else {
	  axval += axes[a].deadzone;
	  if ( axval > 0 ) axval = 0;
	}
	v += axval * motors[m].coefs[a] / 100;
      }
      switch ( motors[m].mode) {
      case 's':  // Speed
	if ( v < -motors[m].maxspeed ) v = -motors[m].maxspeed;
	if ( v >  motors[m].maxspeed ) v =  motors[m].maxspeed;
	fprintf(stderr, "(speed%d=%d)", m, v);
	ev3_motor_speed(ev3, 1<<m, v);
	break;
      case 'p':  // Position
	fprintf(stderr, "(pos%d=%d)", m, v);
	ev3_motor_moveto(ev3, 1<<m, motors[m].gain, motors[m].maxspeed, v);
	break;
      }
    }
    fprintf(stderr, "\n");
    // At least 1000 ms between loops
    usleep(1000);
    // Make sure we send at least one synchronous command,
    // so as not to flood the EV3.
    ev3_output_ready(ev3, MOTA|MOTB|MOTC|MOTD);
  } while ( ! quit_requested );
    
  bthid_close(hid);

  ev3_motor_stop(ev3, MOTA|MOTB|MOTC|MOTD, FREERUN);
  usleep(100000);
  ev3_close(ev3);

  return 0;
}
