/*
 * http://www.pabr.org/bricks/brickhid
 * Bluetooth communication with PS3 controller over L2CAP.
 *
 * Copyright (c) 2014 pabr@pabr.org
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <bluetooth/bluetooth.h>
#include <bluetooth/l2cap.h>

#include "bthid.h"

#define L2CAP_PSM_HIDP_CTRL 0x11
#define L2CAP_PSM_HIDP_INTR 0x13
#define HIDP_TRANS_GET_REPORT    0x40
#define HIDP_TRANS_SET_REPORT    0x50
#define HIDP_DATA_RTYPE_INPUT    0x01
#define HIDP_DATA_RTYPE_OUTPUT   0x02
#define HIDP_DATA_RTYPE_FEATURE  0x03

struct bthid_conn {
  int csk;
  int isk;
};

static void fatal(const char *msg) {
  if ( errno ) perror(msg); else fprintf(stderr, "%s\n", msg);
  exit(1);
}

static int l2cap_listen(const bdaddr_t *bdaddr, unsigned short psm) {
  int sk = socket(PF_BLUETOOTH, SOCK_SEQPACKET, BTPROTO_L2CAP);
  if ( sk < 0 ) fatal("socket");
  struct sockaddr_l2 addr;
  memset(&addr, 0, sizeof(addr));
  addr.l2_family = AF_BLUETOOTH;
  addr.l2_bdaddr = *BDADDR_ANY;
  addr.l2_psm = htobs(psm);
  if ( bind(sk, (struct sockaddr *)&addr, sizeof(addr)) < 0 ) fatal("bind");
  if ( listen(sk,5) < 0 ) fatal("listen");
  return sk;
}

struct bthid_conn *bthid_accept() {
  fprintf(stderr, "Waiting for incoming Bluetooth connection...\n");
  int csk = l2cap_listen(BDADDR_ANY, L2CAP_PSM_HIDP_CTRL);
  int isk = l2cap_listen(BDADDR_ANY, L2CAP_PSM_HIDP_INTR);
  
  struct bthid_conn *conn = (struct bthid_conn *)malloc(sizeof(*conn));
  if ( ! conn ) fatal("malloc");

  conn->csk = accept(csk, NULL, NULL);
  if ( conn->csk < 0 ) fatal("accept");
  conn->isk = accept(isk, NULL, NULL);
  if ( conn->isk < 0 ) fatal("accept");

  close(csk);
  close(isk);

  int flags = fcntl(conn->isk, F_GETFL,0);
  if ( flags < 0 ) fatal("F_GETFL");
  fcntl(conn->isk, F_SETFL, flags|O_NONBLOCK);

  {
    char set03f4[] = { HIDP_TRANS_SET_REPORT | HIDP_DATA_RTYPE_FEATURE, 0xf4,
		       0x42, 0x03, 0x00, 0x00 };
    send(conn->csk, set03f4, sizeof(set03f4), 0);
    unsigned char ack;
    recv(conn->csk, &ack, sizeof(ack), 0);
    #define LED_PERMANENT 0xff, 0x27, 0x00, 0x00, 0x32
    char set0201[] = {
      HIDP_TRANS_SET_REPORT | HIDP_DATA_RTYPE_OUTPUT, 0x01,
      0x00, 0x00, 0x00, 0,0, 0x00, 0x00, 0x00,
      0x00, 1<<1,
      LED_PERMANENT,
      LED_PERMANENT,
      LED_PERMANENT,
      LED_PERMANENT,
      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
    };
    send(conn->csk, set0201, sizeof(set0201), 0);
    recv(conn->csk, &ack, sizeof(ack), 0);
  }

  return conn;
}

void bthid_close(struct bthid_conn *conn) {
  close(conn->csk);
  close(conn->isk);
}

int bthid_recv_axes(struct bthid_conn *conn, int *res, int max_axes) {
  if ( max_axes < PS3_NAXES ) fatal("buffer too small");
  unsigned char report[256];
  int nr = recv(conn->isk, report, sizeof(report), 0);
  if ( nr <= 0 ) return nr;
  if ( nr < 23 ) return 0;
#if 0
  for ( int i=0; i<nr; ++i ) fprintf(stderr, "%02x ", report[i]);
  fprintf(stderr, "\n");
#endif
  res[PS3_SELECT]     = (report[3]&0x01) ? 100 : 0;
  res[PS3_START]      = (report[3]&0x08) ? 100 : 0;
  res[PS3_UP]         = (report[3]&0x10) ? 100 : 0;
  res[PS3_RIGHT]      = (report[3]&0x20) ? 100 : 0;
  res[PS3_DOWN]       = (report[3]&0x40) ? 100 : 0;
  res[PS3_LEFT]       = (report[3]&0x80) ? 100 : 0;
  res[PS3_TRIANGLE]   = (report[4]&0x10) ? 100 : 0;
  res[PS3_CIRCLE]     = (report[4]&0x20) ? 100 : 0;
  res[PS3_CROSS]      = (report[4]&0x40) ? 100 : 0;
  res[PS3_SQUARE]     = (report[4]&0x80) ? 100 : 0;
  res[PS3_PS]         = (report[5]&0x01) ? 100 : 0;
  res[PS3_LH]         = ((int)report[7] - 128 ) * 100 / 128;
  res[PS3_LV]         = (128 - (int)report[8] ) * 100 / 128;
  res[PS3_RH]         = ((int)report[9] - 128 ) * 100 / 128;
  res[PS3_RV]         = (128 - (int)report[10]) * 100 / 128;
  res[PS3_L2]         = (int)report[19] * 100 / 128;
  res[PS3_R2]         = (int)report[20] * 100 / 128;
  res[PS3_L1]         = (int)report[21] * 100 / 128;
  res[PS3_R1]         = (int)report[22] * 100 / 128;
  return PS3_NAXES;
}
