/*
 * http://www.pabr.org/bricks/brickhid
 * Quick and dirty implementation of portions of the EV3 protocol.
 *
 * Copyright (c) 2014 pabr@pabr.org
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "ev3wifi.h"

#define EV3_BROADCAST_PORT 3015
#define EV3_MOTORS 4

#define NOREPLY 0x80  // Reply not needed
#define REPLY   0x00  // Reply expected (useful for debugging)

struct ev3_conn {
  int sock;
  int seq;
};

static void fatal(const char *msg) {
  if ( errno ) perror(msg); else fprintf(stderr, "%s\n", msg);
  exit(1);
}

struct ev3_conn *ev3_connect(const char *localaddr, const char *ev3addr) {
  int res;

  fprintf(stderr, "Waiting for beacon...\n");

  int sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
  if ( sock < 0 ) fatal("socket");

  { // Bind to local port and optional local address
    struct sockaddr_in lsa;
    memset(&lsa, 0, sizeof(lsa));
    lsa.sin_family = AF_INET;
    if ( localaddr )
      if ( ! inet_aton(localaddr, &lsa.sin_addr) ) fatal(localaddr);
    lsa.sin_port = htons(EV3_BROADCAST_PORT);
    res = bind(sock, (struct sockaddr*)&lsa, sizeof(lsa));
    if ( res ) fatal("bind");
  }
  
  // EV3 address filter

  struct in_addr ev3a;
  if ( ev3addr )
    if ( ! inet_aton(ev3addr, &ev3a) ) fatal("ev3addr");

  // Receive beacon messages

  struct sockaddr_in srcsa;
  socklen_t srcsalen;
  int ev3port;
  char ev3sn[32+1];

  while ( 1 ) {
    char msg[256];
    srcsalen = sizeof(srcsa);
    res = recvfrom(sock, msg, sizeof(msg), 0,
		   (struct sockaddr*)&srcsa, &srcsalen);
    if ( res <= 0 ) continue;
    if ( ev3addr && memcmp(&ev3a,&srcsa,srcsalen) ) continue;
    if ( sscanf(msg, "Serial-Number: %32s\nPort: %d\nName: EV3\nProtocol: EV3",
		ev3sn, &ev3port) != 2 ) continue;
    break;
  }

  fprintf(stderr, "Found EV3 %s %d\n", ev3sn, ev3port);

  // Activate TCP on EV3

  char dummy = 0;
  sendto(sock, &dummy, sizeof(dummy), 0, (struct sockaddr*)&srcsa, srcsalen);

  close(sock);

  // Open TCP connection

  fprintf(stderr, "Connecting...\n");

  srcsa.sin_port = htons(ev3port);
  sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
  if ( sock < 0 ) fatal("socket");
  res = connect(sock, (struct sockaddr*)&srcsa, srcsalen);
  if ( res < 0 ) fatal("connect");
  
  // Send magic message

  { char msg[256];
    sprintf(msg, "GET /target?sn=%s VMTP1.0\n", ev3sn);
    write(sock, msg, strlen(msg));
    res = read(sock, msg, 256);
    if ( strncmp(msg, "Accept:EV340",12) ) fatal("Accept");
  }
  
  struct ev3_conn *conn = (struct ev3_conn*)malloc(sizeof(*conn));
  if ( ! conn ) fatal("malloc");

  fprintf(stderr, "Connected.\n");

  conn->sock = sock;
  conn->seq = 0;
  return conn;
}

void ev3_close(struct ev3_conn *conn) {
  close(conn->sock);
}

static int recv_sync(struct ev3_conn *conn, unsigned char *buf, int len) {
  int size, seq;
  do {
    int nr = recv(conn->sock, buf, 2, 0);
    if ( nr != 2 ) fatal("recv");
    size = buf[0] | (buf[1]<<8);
    nr = recv(conn->sock, buf, size, 0);
    if ( nr != size ) fatal("recv2");
#if 0
    fprintf(stderr, "RECV %d:", nr);
    for ( int i=0; i<nr; ++i ) fprintf(stderr, " %02X", buf[i]);
    fprintf(stderr, "\n");
#endif
    seq = buf[0] | (buf[1]<<8);
  } while ( seq != conn->seq );
  return size;
}

int ev3_command(struct ev3_conn *conn, int reply, int var1, int var2,
		unsigned char *command, int len) {
  unsigned char header[7] = {
    5+len, (5+len)>>8,
    conn->seq, conn->seq>>8,
    reply, var1, var2
  };
  struct iovec ios[2] = { {header,sizeof(header)}, {command,len} };
  return writev(conn->sock, ios, 2);
}

// Play tone.

int ev3_sound_tone(struct ev3_conn *conn, int level, int freq, int dur) {
  unsigned char msg[] = { 
    0x94,
    0x01, /*TONE*/
    0x81, level,
    0x82, freq, freq>>8,
    0x82, dur, dur>>8
  };
  ev3_command(conn, NOREPLY,0,0, msg, sizeof(msg));
  ++conn->seq;
  return 0;
}  

// Set speed and start motor.

int ev3_motor_speed(struct ev3_conn *conn, ev3nos nos, ev3s speed) {
  unsigned char msg[] = { 
    0xA5, 0x00, nos,
    0x81, speed
  };
  ev3_command(conn, NOREPLY,0,0, msg, sizeof(msg));
  ++conn->seq;
  return 0;
}

int ev3_motor_start(struct ev3_conn *conn, ev3nos nos) {
  unsigned char msg[] = { 0xA6, 0x00, nos };
  ev3_command(conn, NOREPLY,0,0, msg, sizeof(msg));
  ++conn->seq;
  return 0;
}

int ev3_motor_stop(struct ev3_conn *conn, ev3nos nos, ev3braking brake) {
  unsigned char msg[] = { 0xA3, 0x00, nos, brake };
  ev3_command(conn, NOREPLY,0,0, msg, sizeof(msg));
  ++conn->seq;
  return 0;
}

// Move [cst] ev3p at [speed] (-100%..+100%), with ramp up and down.

int ev3_motor_step_speed(struct ev3_conn *conn, ev3nos nos, ev3s speed,
			 ev3p up, ev3p cst, ev3p down,
			 ev3braking brake) {
  unsigned char msg[] = { 
    0xAE, 0x00, nos,
    0x81, speed,
    0x83, up, up>>8, up>>16, up>>24,
    0x83, cst, cst>>8, cst>>16, cst>>24,
    0x83, down, down>>8, down>>16, down>>24,
    brake
  };
  ev3_command(conn, NOREPLY,0,0, msg, sizeof(msg));
  ++conn->seq;
  return 0;
}

int ev3_motor_set_position(struct ev3_conn *conn, ev3nos nos, ev3p pos) {
  fatal("not implemented");
  unsigned char msg[] = {
    0xAB, 0x00, nos,
    0x83, pos, pos>>8, pos>>16, pos>>24
  };
  ev3_command(conn, REPLY,0,0, msg, sizeof(msg));  // REPLY shows error code
  ++conn->seq;
  return 0;
}

// Reset origin.

int ev3_motor_reset(struct ev3_conn *conn, ev3nos nos) {
  unsigned char msg[] = { 0xA2, 0x00, nos };
  ev3_command(conn, NOREPLY,0,0, msg, sizeof(msg));
  ++conn->seq;
  return 0;
}

// Read speed and position.
// Note: Origin is reset by some commands.

int ev3_motor_read(struct ev3_conn *conn, int port,
		   ev3s *speed, ev3p *pos) {
  unsigned char msg[] = { 
    0xA8, 0x00, port,
    0x40|0x20|0,
    0x40|0x20|4
  };
  ev3_command(conn, REPLY,0x08,0x00, msg, sizeof(msg));
  unsigned char buf[256];
  int len = recv_sync(conn, buf, sizeof(buf));
  if ( len != 11 ) fatal("sync lost");
  *speed = (signed)buf[3];
  *pos = buf[7] | (buf[8]<<8) | (buf[9]<<16) | (buf[10]<<24);
  ++conn->seq;
  return 0;
}

// Reset motor position. Not reset internally.

int ev3_motor_clr_count(struct ev3_conn *conn, ev3nos nos) {
  unsigned char msg[] = { 0xB2, 0x00, nos };
  ev3_command(conn, NOREPLY,0,0, msg, sizeof(msg));
  ++conn->seq;
  return 0;
}

// Read motor position.

int ev3_motor_get_count(struct ev3_conn *conn, int port, ev3p *pos) {
  unsigned char msg[] = { 0xB3, 0x00, port, 0x40|0x20|0 };
  ev3_command(conn, REPLY,0x04,0x00, msg, sizeof(msg));
  unsigned char buf[256];
  int len = recv_sync(conn, buf, sizeof(buf));
  if ( len != 7 ) fatal("sync lost");
  *pos = buf[3] | (buf[4]<<8) | (buf[5]<<16) | (buf[6]<<24);
  ++conn->seq;
  return 0;
}

// Naive servo loop.

int ev3_motor_moveto(struct ev3_conn *conn, ev3nos nos,
		     int gain, ev3s maxspeed, ev3p pos) {
  for ( int no=0; no<EV3_MOTORS; ++no ) {
    if ( ! (nos & (1<<no)) ) continue;
    //ev3s curspeed;
    ev3p curpos;
    //    ev3_motor_read(conn, no, &curspeed, &curpos);
    ev3_motor_get_count(conn, no, &curpos);
    ev3p diff = pos - curpos;
    ev3s speed = diff*diff * gain / 1000;
    if ( diff < 0 ) speed = -speed;
    if ( speed < -maxspeed ) speed = -maxspeed;
    if ( speed >  maxspeed ) speed =  maxspeed;
    fprintf(stderr, "[%d:%d,%d]", no, (int)diff, (int)speed);
    if ( speed ) {
      ev3_motor_speed(conn, 1<<no, speed);
      ev3_motor_start(conn, 1<<no);
    } else
      ev3_motor_stop(conn, 1<<no, BRAKE);
  }
  return 0;
}

int ev3_output_ready(struct ev3_conn *conn, ev3nos nos) {
  unsigned char msg[] = { 0xAA, 0x00, nos };
  ev3_command(conn, REPLY,0,0, msg, sizeof(msg));
  unsigned char buf[256];
  int len = recv_sync(conn, buf, sizeof(buf));
  if ( len != 3 ) fatal("sync lost");
  ++conn->seq;
  return 0;
}
